from loguru import logger
from uuid import uuid4
import asyncio
import tornado
from acedatacloud_scaffold.handlers import HealthHandler
from acedatacloud_scaffold.settings import HTTP_PORT


class BaseController(object):

    def __init__(self):
        self.id = str(uuid4())
        self.application = tornado.web.Application([
            (r"/health", HealthHandler),
        ])

    def add_handler(self, path, handler):
        self.application.add_handlers(
            '.*', [(path, handler)])
        logger.debug(f'add handler {path}')

    async def main(self):
        self.application.listen(HTTP_PORT)
        await asyncio.Event().wait()

    def start(self):
        asyncio.run(self.main())



from rest_framework.exceptions import APIException as BaseAPIException


class APIException(BaseAPIException):

    def __init__(self, detail=None, code=None):
        self.detail = detail or self.default_detail
        self.code = code or self.default_code



from django.utils.translation import gettext_lazy as _
from rest_framework import status
from .base import APIException


class TokenMismatchedException(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _('token and api does not match.')
    default_code = 'token_mismatched'



from django.utils.translation import gettext_lazy as _
from rest_framework import status
from .base import APIException


class NoApiException(APIException):
    status_code = status.HTTP_404_NOT_FOUND
    default_detail = _('API does not exist, please make sure url is correct.')
    default_code = 'no_api'



import environs

env = environs.Env()
env.read_env()

HTTP_HOST = env.str('HTTP_HOST', '0.0.0.0')
HTTP_PORT = env.int('HTTP_PORT', 8000)

DEFAULT_TIMEOUT_FORWARD = env.int('DEFAULT_TIMEOUT_FORWARD', 60)

RECORD_SERVER_URL = env.str(
    'RECORD_SERVER_URL', 'http://platform-record:8000/record')

# error status
ERROR_STATUS_API_ERROR = 500
ERROR_STATUS_UNKNOWN = 500
ERROR_STATUS_CONNECTION_RESET = 499
ERROR_STATUS_NOT_FOUND = 404

# error code
ERROR_CODE_API_ERROR = 'api_error'
ERROR_CODE_UNKNOWN = 'unknown'
ERROR_CODE_CONNECTION_RESET = 'connection_reset'
ERROR_CODE_NOT_FOUND = 'not_found'

# error detail
ERROR_DETAIL_API_ERROR = 'api internal error, please contact admin'
ERROR_DETAIL_UNKNOWN = 'unknown internal server error, please contact admin'
ERROR_DETAIL_CONNECTION_RESET = 'connection reset error, this usually caused by client closed connection'
ERROR_DETAIL_NOT_FOUND = 'not found'



from .base import BaseHandler
from .health import HealthHandler
from .callback import CallbackHandler
from .hybrid import HybridHandler


__all__ = [
    'BaseHandler',
    'HealthHandler',
    'CallbackHandler',
    'HybridHandler',
]



from rest_framework import status
from .base import APIException


class TooManyRequestsException(APIException):
    status_code = status.HTTP_429_TOO_MANY_REQUESTS
    default_code = 'too_many_requests'



from .controllers import BaseController
from .handlers import BaseHandler
from .handlers.callback import CallbackHandler
from .handlers.health import HealthHandler
from .handlers.hybrid import HybridHandler


__all__ = [
    'BaseController',
    'BaseHandler',
    'CallbackHandler',
    'HealthHandler',
    'HybridHandler',
]



from loguru import logger


class LogMixin(object):

    def log(self, level, message):
        """
        Custom log method that includes the trace ID.
        """
        if isinstance(message, bytes):
            message = message.decode('utf-8')
        trace_message = f"[{self.trace_id}] {message}"
        getattr(logger, level)(trace_message)

    @property
    def logger(self):
        """
        Expose a logger-like interface that automatically includes the trace ID.
        """
        class LoggerAdapter:
            def __init__(self, outer):
                self.outer = outer

            def debug(self, message):
                self.outer.log('debug', message)

            def info(self, message):
                self.outer.log('info', message)

            def warning(self, message):
                self.outer.log('warning', message)

            def error(self, message):
                self.outer.log('error', message)

            def critical(self, message):
                self.outer.log('critical', message)

            def exception(self, message):
                self.outer.log('exception', message)

        return LoggerAdapter(self)



import httpx
from acedatacloud_scaffold.settings import DEFAULT_TIMEOUT_FORWARD


class ForwardMixin(object):

    async def get_forward_request_timeout(self):
        return httpx.Timeout(DEFAULT_TIMEOUT_FORWARD, read=None)

    async def get_forward_request_method(self):
        return self.request.method

    async def get_forward_request_url(self):
        return self.request.uri

    async def get_forward_request_headers(self):
        return self.request.headers

    async def get_forward_request_body(self):
        return self.request.body

    async def get_forward_response_status(self):
        return self.forward_response.status_code

    async def get_forward_response_headers(self):
        return self.forward_response.headers

    async def get_forward_response_body(self):
        async for data in self.forward_response.aiter_raw():
            yield data

    async def write_response_body(self):
        async for data in self.get_forward_response_body():
            self.logger.debug(
                f'write data {data}')
            self.write(data)
        self.logger.debug('finish write data')
        self.finish()

    async def write_response_status(self):
        forward_response_status = await self.get_forward_response_status()
        self.set_status(forward_response_status)

    async def write_response_headers(self):
        # set response headers
        forward_response_headers = await self.get_forward_response_headers()
        for header, value in forward_response_headers.items():
            self.logger.debug(
                f'set header {header} {value}')
            self.set_header(header, value)

    async def forward(self):
        # forward info for building forward request
        forward_timeout = await self.get_forward_request_timeout()
        forward_url = await self.get_forward_request_url()
        forward_headers = await self.get_forward_request_headers()
        forward_body = await self.get_forward_request_body()
        forward_method = await self.get_forward_request_method()

        async with httpx.AsyncClient(timeout=forward_timeout) as client:
            self.logger.debug(
                f'forward_url {forward_url} forward_headers {forward_headers} forward_body {forward_body}')
            async with client.stream(
                forward_method,
                forward_url,
                headers=forward_headers,
                data=forward_body,
            ) as response:
                self.forward_response = response
                self.logger.debug(
                    f'forward response {self.forward_response}')
                await self.write_response_status()
                await self.write_response_headers()
                await self.write_response_body()



from django.utils.translation import gettext_lazy as _
from rest_framework import status
from .base import APIException


class InvalidTokenException(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _(
        'The token is invalid, please make sure your token is correct.')
    default_code = 'invalid_token'



from .base import APIException
from .bad_request import BadRequestException
from .too_many_requests import TooManyRequestsException
from .no_token import NoTokenException
from .invalid_token import InvalidTokenException
from .used_up import UsedUpException
from .no_api import NoApiException
from .token_mismatched import TokenMismatchedException
from .api_not_implemented import ApiNotImplementedException


__all__ = [
    'APIException',
    'BadRequestException',
    'TooManyRequestsException',
    'NoTokenException',
    'InvalidTokenException',
    'UsedUpException',
    'NoApiException',
    'TokenMismatchedException',
    'ApiNotImplementedException',
]



from django.utils.translation import gettext_lazy as _
from rest_framework import status
from .base import APIException


class UsedUpException(APIException):
    status_code = status.HTTP_403_FORBIDDEN
    default_detail = _(
        'Your packaged has been used up, please buy more in system.')
    default_code = 'used_up'



from loguru import logger
from acedatacloud_scaffold.exceptions import APIException
import json
from acedatacloud_scaffold.settings import \
    ERROR_STATUS_API_ERROR, ERROR_CODE_API_ERROR, RECORD_SERVER_URL
from tornado.web import RequestHandler
from acedatacloud_scaffold.mixins import LogMixin
from uuid import uuid4
import requests


class BaseHandler(RequestHandler, LogMixin):

    def write_json(self, data, status=200, headers={}):
        self.set_status(status)
        self.set_header('Content-Type', 'application/json')
        for key, value in headers.items():
            self.set_header(key, value)
        self.write(json.dumps(data, ensure_ascii=False))
        self.logger.debug(f'write json {data}')

    def write_success_sync(self, data, status=200, headers={}):
        self.write_json(data, status=status, headers=headers)
        self.finish()

    def write_success(self, data, status=200, headers={}):
        self.write_success_sync(data, status=status, headers=headers)

    def write_error(self, status_code, **kwargs):
        self.write_error_sync(status_code, **kwargs)

    def write_error_sync(self, status_code, **kwargs):
        self.logger.error(f'error happened {kwargs}')
        exception = None
        if "exc_info" in kwargs:
            exception = kwargs["exc_info"][1]
        self.logger.exception(f'error {exception}')
        data = {}
        status = status_code or exception.status_code if hasattr(
            exception, 'status_code') else ERROR_STATUS_API_ERROR
        self.logger.debug(f'{self.trace_id} error status {status}')
        # construct error
        data = {
            'trace_id': self.trace_id,
            'error': {
                'code': (exception.code if hasattr(exception, 'code') else
                         exception.default_code) if isinstance(exception, APIException) else ERROR_CODE_API_ERROR,
                'message': exception.detail if isinstance(exception, APIException) else str(exception),
            }
        }
        self.write_json(data, status=status)
        self.logger.debug(f'write error {data}')
        self.finish()

    @logger.catch
    def record(self, data, extra_data={}):
        record_server_url = RECORD_SERVER_URL
        self.logger.debug(f'record url {record_server_url}')
        data = {
            'status': data.get('status'),
            'trace_id': self.trace_id,
            'application_id': self.application_id,
            'metadata': {
                'task_id': self.task_id,
            },
            'request': {
                'body': self.request.body.decode('utf-8'),
                'json': json.loads(self.request.body),
                'query': self.request.query_arguments,
                'headers': dict(self.request.headers),
            }
        }
        data.update(extra_data)
        logger.debug(f'{self.trace_id} record data {data}')
        response = requests.post(record_server_url, json=data)
        logger.debug(f'{self.trace_id} record response {response}')

    def initialize_trace_id(self):
        trace_id = self.request.query_arguments.get('trace_id')
        self.trace_id = trace_id[0].decode(
            'utf-8') if trace_id and len(trace_id) > 0 else str(uuid4())
        logger.debug(f'trace id {self.trace_id}')

    def initialize_application_id(self):
        application_id = self.request.query_arguments.get('application_id')
        self.application_id = application_id[0].decode(
            'utf-8') if application_id and len(application_id) > 0 else None
        logger.debug(f'application id {self.application_id}')

    def initialize_task_id(self):
        self.task_id = str(uuid4())
        logger.debug(f'task id {self.task_id}')

    def initialize(self):
        self.initialize_trace_id()
        self.initialize_task_id()
        self.initialize_application_id()



from django.utils.translation import gettext_lazy as _
from rest_framework import status
from .base import APIException


class NoTokenException(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _('No token specified in url query.')
    default_code = 'no_token'



from acedatacloud_scaffold.handlers.base import BaseHandler
import json
from loguru import logger
import requests
from acedatacloud_scaffold.settings import \
    ERROR_STATUS_API_ERROR, ERROR_CODE_API_ERROR
from acedatacloud_scaffold.exceptions import APIException, BadRequestException


class CallbackHandler(BaseHandler):

    def initialize(self):
        super().initialize()
        self.initialize_async_context()

    @logger.catch
    def send_callback(self, data):
        self.logger.debug(f'{self.trace_id} callback url {self.callback_url}')
        self.logger.debug(f'{self.trace_id} callback data {data}')
        response = requests.post(self.callback_url, json=data)
        self.logger.debug(f'{self.trace_id} callback response {response}')

    def write_success_async(self, data, status=200, headers={}):
        # append task id
        data['task_id'] = self.task_id
        # append trace id
        data['trace_id'] = self.trace_id
        self.send_callback(data)

    def write_index(self):
        self.write_json({
            'task_id': self.task_id,
        })
        self.finish()
        self._finished = False
        # this flag used for checking if index has been written,
        # if index not written, we can return error sync
        self.index_written = True

    def write_success(self, data, status=200, headers={}):
        self.write_success_async(data, status=status, headers=headers)

    def write_error(self, status_code, **kwargs):
        self.write_error_async(status_code, **kwargs)
        if not self.index_written:
            self.write_error_sync(status_code, **kwargs)

    def write_error_async(self, status_code, **kwargs):
        self.logger.error(f'error happened {kwargs}')
        exception = None
        if "exc_info" in kwargs:
            exception = kwargs["exc_info"][1]
        self.logger.exception(f'error {exception}')
        data = {}
        status = status_code or exception.status_code if hasattr(
            exception, 'status_code') else ERROR_STATUS_API_ERROR
        self.logger.debug(f'{self.trace_id} error status {status}')
        # construct error
        data = {
            'task_id': self.task_id,
            'trace_id': self.trace_id,
            'error': {
                'code': (exception.code if hasattr(exception, 'code') else
                         exception.default_code) if isinstance(exception, APIException) else ERROR_CODE_API_ERROR,
                'message': exception.detail if isinstance(exception, APIException) else str(exception),
            }
        }
        self.send_callback(data)
        self.logger.debug(f'write error {data}')

    def initialize_async_context(self):
        try:
            request_body = self.request.body.decode('utf-8')
            post_json = json.loads(request_body)
            callback_url = post_json.get('callback_url')
            if not callback_url:
                raise BadRequestException('callback_url is required')
            self.index_written = False
        except Exception:
            pass



from acedatacloud_scaffold.handlers.callback import CallbackHandler
import json
from acedatacloud_scaffold.exceptions import BadRequestException


class HybridHandler(CallbackHandler):

    def initialize_async_context(self):
        try:
            request_body = self.request.body.decode('utf-8')
            post_json = json.loads(request_body)
            callback_url = post_json.get('callback_url')
            if callback_url:
                self.async_mode = True
                self.callback_url = callback_url
                self.index_written = False
            else:
                self.async_mode = False
                self.callback_url = None
        except Exception:
            self.logger.error('error initializing async context')
            raise BadRequestException(
                'error when parsing request body for callback url')

    def write_success(self, data, status=200, headers={}):
        if self.async_mode:
            self.write_success_async(data, status=status, headers=headers)
        else:
            self.write_success_sync(data, status=status, headers=headers)

    def write_error(self, status_code, **kwargs):
        if self.async_mode:
            self.write_error_async(status_code, **kwargs)
            if not self.index_written:
                self.write_error_sync(status_code, **kwargs)
        else:
            self.write_error_sync(status_code, **kwargs)



from .base import APIException
from django.utils.translation import gettext_lazy as _
from rest_framework import status


class ApiNotImplementedException(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _('The API is not implemented.')
    default_code = 'api_not_implemented'
